package xml;

import membranModel.MembranModel;
import membranModel.MembranModelSingleton;

import org.jdom.*;
import org.jdom.input.*;

import pdbParser.PDBreader;
import protein.Protein;

import java.util.*;
import java.awt.Color;
import java.io.IOException;

import javax.vecmath.Matrix3f;

public class XMLReader {

	/**
	 * New contructor.
	 * Executes the readingXML function.
	 * 
	 * @param path Path of the file that should be parsed
	 * @throws IOException
	 * @throws JDOMException
	 */
	public XMLReader(String path) throws IOException, JDOMException {
		MembranModel membranModel = MembranModelSingleton.getInstance();
		membranModel.getHash().clear();
		readingXML(path, membranModel); 
	}
	
	/**
	 * Reads the XML file using the SAXBuilder()
	 * 
	 * @param path Path of the XML file
	 * @param membranModel MembranModel object
	 * @throws IOException
	 * @throws JDOMException
	 */
	public void readingXML(String path, MembranModel membranModel) throws IOException,JDOMException {
		Document doc = new SAXBuilder().build(path);
		Element description = doc.getRootElement();
		
		membranModel.setName(description.getAttribute("name").getValue());
		membranModel.setDescription(description.getAttribute("description").getValue());
		membranModel.setSpecies(description.getAttribute("species").getValue().split(";"));
		membranModel.setCelltype(description.getAttribute("celltype").getValue());
		membranModel.setDate(description.getAttribute("date").getValue());
		
		Element surfaceArea = description.getChild("surfaceArea");
		membranModel.setVRMLFile(surfaceArea.getAttribute("filename").getValue());
		
		PDBreader pdbreader = new PDBreader();
		
		Element lipids = description.getChild("lipids");
		List lipidList = lipids.getChildren("lipidcomponent");	   
		for(int i=0; i<lipidList.size(); i++) {
			membranModel.setBiLayerType(((Element)lipidList.get(i)).getAttribute("pdbcode").getValue());
			membranModel.getBiLayerType().setFilepath((((Element)lipidList.get(i)).getAttribute("pdbfile").getValue()));
			membranModel.getBiLayerType().setName(((((Element)lipidList.get(i)).getAttribute("name").getValue())));
			membranModel.getBiLayerType().startup();
		}
		
		Element proteins = description.getChild("proteins");
		List proteinList = proteins.getChildren("protein");
		
		for(int i=0; i<proteinList.size(); i++) {
			Protein protein = pdbreader.getProtein(null,((Element)proteinList.get(i)).getAttribute("pdbfile").getValue());
			protein.setSize2(convertStringToDoubleArray(((Element)proteinList.get(i)).getAttribute("size").getValue(),"x"));
			protein.setYTranslationLoad(Float.parseFloat(((Element)proteinList.get(i)).getAttribute("ytanslation").getValue()));
			protein.setRatio(Float.parseFloat(((Element)proteinList.get(i)).getAttribute("ratio").getValue()));
			protein.setQuantity(Integer.parseInt(((Element)proteinList.get(i)).getAttribute("quantity").getValue()));
			protein.setRotationMatrix(convertStringToMatrix3f(((Element)proteinList.get(i)).getAttribute("rotationMatrix").getValue(),"x"));
			protein.setComponent(((Element)proteinList.get(i)).getAttribute("component").getValue());
			protein.setSubcomponent(((Element)proteinList.get(i)).getAttribute("subcomponent").getValue());
			protein.setModified(((Element)proteinList.get(i)).getAttribute("manipulated").getValue().equals("true"));
			protein.setSelfmade(((Element)proteinList.get(i)).getAttribute("usergenerated").getValue().equals("true"));
			int[] tmpC = convertStringToIntArray(((Element)proteinList.get(i)).getAttribute("color").getValue(),"x");
			protein.setProteinColor(new Color(tmpC[0],tmpC[1],tmpC[2]));
			Vector tmpP = new Vector();
			String[] tmpS = convertStringToStringArray(((Element)proteinList.get(i)).getAttribute("positions").getValue(),"_");
			for (int j=0; j<tmpS.length; j++) {
				tmpP.add(convertStringToIntArray(tmpS[j],";"));
			}
			protein.setProteinPositions(tmpP);	
			protein.setNodeName(((Element)proteinList.get(i)).getAttribute("nodename").getValue());
			membranModel.addProtein(protein);
		}
	}	
	
	/**
	 * Converts a string into an array of int values. 
	 * 
	 * @param string String
	 * @param separator Separator of the values in the given string
	 * @return int[]
	 */
	public int[] convertStringToIntArray(String string, String separator) {
		StringTokenizer tokenizer = new StringTokenizer(string, separator);
		int output[] = new int[tokenizer.countTokens()];		    
		for (int i = 0; tokenizer.hasMoreTokens(); ++i) {
		  output[i] = Integer.parseInt(tokenizer.nextToken());
		}
		return output;
	}
	
	/**
	 * Converts a string into a Matrix3f.
	 * 
	 * @param string String
	 * @param separator Separator of the values in the given string
	 * @return Matrix3f
	 */
	public Matrix3f convertStringToMatrix3f(String string, String separator) {
		StringTokenizer tokenizer = new StringTokenizer(string, separator);
		float tmp[] = new float[tokenizer.countTokens()];		    
		for (int i = 0; tokenizer.hasMoreTokens(); ++i) {
		  tmp[i] = Float.parseFloat(tokenizer.nextToken());
		}
		Matrix3f matrix = new Matrix3f();
		matrix.m00 = tmp[0];
		matrix.m01 = tmp[1];
		matrix.m02 = tmp[2];
		matrix.m10 = tmp[3];
		matrix.m11 = tmp[4];
		matrix.m12 = tmp[5];
		matrix.m20 = tmp[6];
		matrix.m21 = tmp[7];
		matrix.m22 = tmp[8];
		return matrix;
	}
		
	/**
	 * Converts a string into an array of double values.
	 * 
	 * @param string String
	 * @param separator Separator of the values in the given string
	 * @return double[]
	 */
	public double[] convertStringToDoubleArray(String string, String separator) {
		StringTokenizer tokenizer = new StringTokenizer(string, separator);
		double output[] = new double[tokenizer.countTokens()];		    
		for (int i = 0; tokenizer.hasMoreTokens(); ++i) {
		  output[i] = Double.parseDouble(tokenizer.nextToken());
		}
		return output;
	}
	
	/**
	 * Converts a string into an array of strings.
	 * 
	 * @param string String
	 * @param separator Separator of the values in the given string
	 * @return String[]
	 */
	public String[] convertStringToStringArray(String string, String separator) {
		StringTokenizer tokenizer = new StringTokenizer(string, separator);
		String output[] = new String[tokenizer.countTokens()];		    
		for (int i = 0; tokenizer.hasMoreTokens(); ++i) {
		  output[i] = tokenizer.nextToken();
		}
		return output;
	}
}
