package xml;

import membranModel.MembranModel;

import org.jdom.Element;
import org.jdom.Document;
import org.jdom.output.XMLOutputter;
import org.jdom.output.Format;
import protein.Protein;

import java.io.IOException;
import java.io.FileOutputStream;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.vecmath.Matrix3f;

public class XMLWriter {	

	String path;
	
	/**
	 * Empty constructor.
	 *
	 */
	public XMLWriter() {}
	
	/**
	 * New constructor.
	 * 
	 * @param path Path where the XML file should be written
	 * @param membranModel MembranModel object
	 */
	public XMLWriter(String path, MembranModel membranModel) {
		this.path = path;
		writer_start(this.path, membranModel);
	}
	
	/**
	  * Creates a new document using the createFile and writeFile functions.
	  * 
	  * @param filename file name
	  * @param membranModel MembranModel object
	  */
	public void writer_start(String filename, MembranModel membranModel) {
		Document doc = createFile(filename, membranModel);
		
		try {
			writeFile(doc, filename);
		} catch (IOException e) {
			JOptionPane.showMessageDialog(null,
				  "Error occured during write the file!" +
				  "\n\n" +
				  "Please try again.",
				  "Write Error",
				   JOptionPane.WARNING_MESSAGE);
		}
	}
	
	/**
	 * Parse the data of the MembranModel object and create an document.
	 * 
	 * @param filename File name
	 * @return document Document which contains the MembranModel data
	 */
	public Document createFile(String filename, MembranModel membranModel) {
		Document doc = new Document();
		
		Element rootElement = new Element("MEMBRANE-description");
		doc.setRootElement(rootElement);
		rootElement.setAttribute("name", membranModel.getName()+"");
		rootElement.setAttribute("description", membranModel.getDescription()+"");
		rootElement.setAttribute("species", convertStringArrayToString(membranModel.getSpecies(),";"));
		rootElement.setAttribute("celltype", membranModel.getCelltype()+"");
		rootElement.setAttribute("date", membranModel.getDate()+"");
		
		Element surface = new Element("surfaceArea");
		surface.setAttribute("filename",membranModel.getVRMLFile()+"");
		rootElement.addContent(surface);
		
		Protein tmpLipid = membranModel.getBiLayerType();
		Element lipid = new Element("lipids");		
		Element lipidcomponent = new Element("lipidcomponent");
		lipidcomponent.setAttribute("pdbcode",tmpLipid.getPDBcode()+"");
		lipidcomponent.setAttribute("pdbfile",tmpLipid.getFilepath()+"");
		lipidcomponent.setAttribute("name",tmpLipid.getName()+"");
		lipid.addContent(lipidcomponent);
		rootElement.addContent(lipid);
		
		Vector membranProteinVector = membranModel.getAllProteins();
		Protein tmpProtein;
		Element proteins = new Element("proteins");
		for (int i=0; i<membranProteinVector.size(); i++) {
			Element protein = new Element("protein");
			tmpProtein = (Protein)membranProteinVector.get(i);
			protein.setAttribute("name",tmpProtein.getName()+"");
			protein.setAttribute("ratio",tmpProtein.getRatio()+"");
			protein.setAttribute("quantity",tmpProtein.getQuantity()+"");
			protein.setAttribute("pdbfile",tmpProtein.getFilepath());
			protein.setAttribute("pdbcode",tmpProtein.getPDBcode()+"");
			protein.setAttribute("size",convertDoubleArrayToString(tmpProtein.getSize(),"x"));
			protein.setAttribute("ytanslation", tmpProtein.getYTranslation()+"");
			protein.setAttribute("rotationMatrix",convertMatrix3fToString(tmpProtein.getRotationMatrix(),"x"));
			protein.setAttribute("component",tmpProtein.getComponent()+"");
			protein.setAttribute("subcomponent",tmpProtein.getSubcomponent()+"");
			protein.setAttribute("manipulated",tmpProtein.isModified()+"");
			protein.setAttribute("usergenerated",tmpProtein.isSelfmade()+"");
			protein.setAttribute("color",tmpProtein.getProteinColor().getRed()+"x"+tmpProtein.getProteinColor().getGreen()+"x"+tmpProtein.getProteinColor().getBlue());
			String tmpPos = "";
			for (int j=0; j<tmpProtein.getProteinPositions().size(); j++) {				
				tmpPos += ((int[])tmpProtein.getProteinPositions().get(j))[0]+";"+((int[])tmpProtein.getProteinPositions().get(j))[1]+";"+((int[])tmpProtein.getProteinPositions().get(j))[2];
				tmpPos += (j < (tmpProtein.getProteinPositions().size() - 1)) ? "_" : ""; 
			}
			protein.setAttribute("positions",tmpPos);
			protein.setAttribute("nodename", tmpProtein.getNodeName()+"");
			proteins.addContent(protein);
		}
		
		rootElement.addContent(proteins);
		
		return doc;
	}

	/**
	 * Write the XML file with the XMLOutputter class.
	 * 
	 * @param doc Document
	 * @param filename File name
	 * @throws IOException
	 */
	public void writeFile(Document doc, String filename) throws IOException {
		XMLOutputter outputter = new XMLOutputter();
		outputter.setFormat(Format.getPrettyFormat());
		FileOutputStream output = new FileOutputStream(filename);
		outputter.output(doc, output);
	}
	
	/**
	 * Converts an int array into a string.
	 * 
	 * @param array Array of int values
	 * @param separator Separator for the values in the generated string
	 * @return String
	 */
	public String convertIntArrayToString(int[] array, String separator) {
		String output = "";
		for (int i=0; i<array.length; i++) {
			output += (i < (array.length-1)) ? array[i]+separator : array[i]+"";
		}		
		return output;
	}
	
	/**
	 * Converts a Matrix3f into a string.
	 * 
	 * @param matrix Matrix3f
	 * @param separator Separator for the values in the generated string
	 * @return String
	 */
	public String convertMatrix3fToString(Matrix3f matrix, String separator) {
		String output = "";
		output += matrix.m00+separator+matrix.m01+separator+matrix.m02+separator+
				  matrix.m10+separator+matrix.m11+separator+matrix.m12+separator+
				  matrix.m20+separator+matrix.m21+separator+matrix.m22;
		return output;
	}
	
	/**
	 * Converts a vector of int arrays into a string.
	 * 
	 * @param vector Vector of int arrays of length 3.
	 * @param separator Separator for the values in the generated string
	 * @return String
	 */
	public String convertVectorIntArrayToString(Vector vector, String separator) {
		String output = "";
		for (int i=0; i<vector.size(); i++) {
			output += ((int[])vector.get(i))[0] + separator;
			output += ((int[])vector.get(i))[1] + separator;
			output += ((int[])vector.get(i))[2];
			output += (i < (vector.size()-1)) ? ";" : "";
		}		
		return output;
	}
	
	/**
	 * Converts an array of double values into a string.
	 * 
	 * @param array Array of double values
	 * @param separator Separator for the values in the generated string
	 * @return String
	 */
	public String convertDoubleArrayToString(double[] array, String separator) {
		String output = "";
		for (int i=0; i<array.length; i++) {
			output += (i < (array.length-1)) ? array[i]+separator : array[i]+"";
		}		
		return output;
	}
	
	/**
	 * Converts an array of strings into a single string.
	 * 
	 * @param array Array of strings
	 * @param separator Separator for the values in the generated string
	 * @return String
	 */
	public String convertStringArrayToString(String[] array, String separator) {
		String output = "";
		for (int i=0; i<array.length; i++) {
			output += (i < (array.length-1)) ? array[i]+separator : array[i];
		}		
		return output;
	}
	
	/**
	 * @return path String (Path of the file that should be written.)
	 */
	public String getPath() {
		return path;
	}

	/**
	 * @param string String that contains the path where the file should be written.
	 */
	public void setPath(String string) {
		path = string;
	}	
}
